import marimo

__generated_with = "0.14.9"
app = marimo.App(width="full")


@app.cell
def _(mo):
    mo.md(
        r"""
    ## Spectrometer calibration tool

    First, select the calibration lamp as measured by the spectrometer you want to calibrate.
    """
    )
    return


@app.cell
def _(mo):
    measurement_browser = mo.ui.file_browser(multiple=False)
    measurement_browser
    return (measurement_browser,)


@app.cell
def _(mo):
    mo.md(r"""Next, select the calibration lamp reference file from the list:""")
    return


@app.cell
def _(aw, mo):
    lamp_options = mo.ui.dropdown(
        options=[e.value for e in aw.spectrum.CalibrationLampReferences]
    )
    lamp_options
    return (lamp_options,)


@app.cell
def _(aw, measurement_browser, mo):
    _path = measurement_browser.path()
    mo.stop(_path is None)
    direct = aw.data.load_mean_spectrum_from_scarab(_path).to_normalized()
    return (direct,)


@app.cell
def _(aw, lamp_options, mo):
    mo.stop(lamp_options.value is None)
    reference = aw.data.load_mean_spectrum_from_scarab(
        aw.spectrum.get_calibration_path() / lamp_options.value
    ).to_normalized()
    return (reference,)


@app.cell
def _(aw, np, scipy):
    def generate_response_curve(
        wavelength: np.ndarray, coefficients: np.ndarray
    ) -> np.ndarray:
        relative_wl = wavelength - coefficients[0]
        taylor_series = coefficients[1] + 0.5 * relative_wl * coefficients[2]
        gaussian = np.exp(
            -(np.abs(relative_wl) ** coefficients[4])
            / (2 * coefficients[3] ** coefficients[4])
        )
        return taylor_series * gaussian

    def get_new_wavelength(wavelengths_micron, taylor_coefficients_micron):
        distance = wavelengths_micron - taylor_coefficients_micron[0]
        taylor_shift = taylor_coefficients_micron[1] * np.ones(
            distance.shape, dtype=float
        )
        for i in range(2, len(taylor_coefficients_micron)):
            taylor_shift += distance ** (i - 1) * taylor_coefficients_micron[i]

        new_wavelengths = 1e-6 * (wavelengths_micron + taylor_shift)
        new_freq = scipy.constants.speed_of_light / new_wavelengths
        return new_wavelengths, new_freq

    def fit_calibration_amplitude_model(
        measurement, reference, wavelength_coeffs, amplitude_guess, roi
    ):
        initial_cal = generate_calibration_from_coeffs(
            amplitude_guess, wavelength_coeffs, measurement.wavelength
        )
        reference_shift = reference.to_interpolated_wavelength(
            initial_cal.corrected_wavelengths
        )

        def residual_amp(coeff):
            cal = generate_calibration_from_coeffs(
                coeff, wavelength_coeffs, measurement.wavelength
            )
            calibrated = cal.apply_to_spectrum(measurement)
            residuals = calibrated.spectrum - reference_shift.spectrum
            return residuals[
                (calibrated.wavelength > roi[0]) & (calibrated.wavelength < roi[1])
            ]

        amplitude_result = scipy.optimize.least_squares(
            residual_amp, amplitude_guess, ftol=1e-12, max_nfev=16384
        )

        return amplitude_result.x

    def generate_calibration_from_coeffs(
        amplitude_coeffs, wavelength_coeffs, wavelengths
    ):
        new_wavelengths, new_freqs = get_new_wavelength(
            wavelengths * 1e6, wavelength_coeffs
        )
        intensity_factors = 1.0 / generate_response_curve(
            wavelength=new_wavelengths * 1e6, coefficients=amplitude_coeffs
        )
        return aw.data.SpectrometerCalibration(
            intensity_factors=intensity_factors,
            corrected_wavelengths=new_wavelengths,
            corrected_frequencies=new_freqs,
        )

    return fit_calibration_amplitude_model, generate_calibration_from_coeffs


@app.cell
def _(mo):
    mo.md(r"""Now, manually adjust the wavelength axis (optional - if it's not necessary, set offset and slope to zero)""")
    return


@app.cell
def _(mo):
    wavelength0 = mo.ui.slider(
        value=0.75,
        start=0,
        stop=3,
        step=1e-4,
        label="Wavelength shift center",
        show_value=True,
    )
    wavelength_offset = mo.ui.slider(
        value=-0.0107,
        start=-1.0,
        stop=1.0,
        step=1e-4,
        label="Wavelength offset",
        show_value=True,
    )
    wavelength_slope = mo.ui.slider(
        value=2e-3,
        start=-1,
        stop=1,
        step=1e-3,
        label="Wavelength slope",
        show_value=True,
    )
    mo.output.append(wavelength0)
    mo.output.append(wavelength_offset)
    mo.output.append(wavelength_slope)
    return wavelength0, wavelength_offset, wavelength_slope


@app.cell
def _(mo):
    mo.md(
        r"""
    Next, set the parameters of the response model. It is a slanted supergaussian. Minimize the difference between the calibrated measurement and the reference to give the fitting routine a good initial guess.

    Once it's decent, click run fitting.

    The "noise level" parameter will affect how the weights are adjusted based on the residuals of this fitting. If the noise level is zero, the weights will be set such that the calibrated spectrum exactly matches the reference, even for values where the measurement signal was very low. If the noise level is increased, the correction to the weights will be reduced for low signal-to-noise wavelengths.
    """
    )
    return


@app.cell
def _(mo):
    amplitude_lam0 = mo.ui.slider(
        value=0.65,
        start=0,
        stop=3,
        step=1e-4,
        label="Amplitude correction center",
        show_value=True,
    )
    amplitude_offset = mo.ui.slider(
        value=3.0,
        start=0,
        stop=30,
        step=1e-4,
        label="Amplitude multiplier",
        show_value=True,
    )
    amplitude_slope = mo.ui.slider(
        value=4.0,
        start=-30,
        stop=30,
        step=1e-4,
        label="Amplitude slope",
        show_value=True,
    )
    amplitude_width = mo.ui.slider(
        value=0.32, start=0, stop=3, step=1e-4, label="Amplitude width", show_value=True
    )
    amplitude_order = mo.ui.slider(
        value=4.0,
        start=2,
        stop=32,
        step=1e-4,
        label="Amplitude Gaussian order",
        show_value=True,
    )
    wiener_noise_level = mo.ui.slider(
        value=0.01, start=0.0, stop=0.1, step=1e-4, label="Noise level", show_value=True
    )
    fit_button = mo.ui.run_button(label="Run fitting")
    mo.output.append(amplitude_lam0)
    mo.output.append(amplitude_offset)
    mo.output.append(amplitude_slope)
    mo.output.append(amplitude_width)
    mo.output.append(amplitude_order)
    mo.output.append(wiener_noise_level)
    mo.output.append(fit_button)

    return (
        amplitude_lam0,
        amplitude_offset,
        amplitude_order,
        amplitude_slope,
        amplitude_width,
        fit_button,
        wiener_noise_level,
    )


@app.cell
def _(
    amplitude_lam0,
    amplitude_offset,
    amplitude_order,
    amplitude_slope,
    amplitude_width,
    direct,
    generate_calibration_from_coeffs,
    np,
    wavelength0,
    wavelength_offset,
    wavelength_slope,
):
    initial_guess = np.array(
        [
            amplitude_lam0.value,
            amplitude_offset.value,
            amplitude_slope.value,
            amplitude_width.value,
            amplitude_order.value,
        ]
    )
    wavelength_guess = [
        wavelength0.value,
        wavelength_offset.value,
        wavelength_slope.value,
    ]
    guess_cal = generate_calibration_from_coeffs(
        initial_guess, wavelength_guess, direct.wavelength
    )
    first_guess = guess_cal.apply_to_spectrum(direct)

    return first_guess, initial_guess, wavelength_guess


@app.cell
def _(
    aw,
    direct,
    first_guess,
    fit_button,
    fit_calibration_amplitude_model,
    generate_calibration_from_coeffs,
    initial_guess,
    np,
    plt,
    reference,
    wavelength_guess,
    wiener_noise_level,
):
    fig, ax = plt.subplots(2, 3, figsize=(16, 9))
    print(ax.shape)
    ax[0, 0].plot(direct.wavelength_nm(), direct.spectrum, label="Measurement")
    ax[0, 0].plot(reference.wavelength_nm(), reference.spectrum, label="Reference")
    ax[0, 0].set_xlabel("Wavelength (nm)")
    ax[0, 0].set_ylabel("Intensity (Arb. unit)")
    ax[0, 0].legend()
    ax[0, 1].plot(
        first_guess.wavelength_nm(), first_guess.spectrum, label="Initial guess"
    )
    ax[0, 1].plot(reference.wavelength_nm(), reference.spectrum, label="Reference")
    ax[0, 1].set_ylabel("Intensity (Arb. unit)")
    ax[0, 1].set_xlabel("Wavelength (nm)")
    ax[0, 1].legend()
    if fit_button.value:
        new_guess = fit_calibration_amplitude_model(
            measurement=direct,
            reference=reference,
            wavelength_coeffs=wavelength_guess,
            amplitude_guess=initial_guess,
            roi=np.array([200e-9, 950e-9]),
        )
        parametrized_cal = generate_calibration_from_coeffs(
            new_guess, wavelength_guess, direct.wavelength
        )
        calibrated = parametrized_cal.apply_to_spectrum(direct)

        ax[0, 2].plot(
            calibrated.wavelength_nm(), calibrated.spectrum, label="Model calibrated"
        )
        ax[0, 2].plot(reference.wavelength_nm(), reference.spectrum, label="Reference")
        ax[0, 2].set_ylabel("Intensity (Arb. unit)")
        ax[0, 2].set_xlabel("Wavelength (nm)")
        ax[0, 2].legend()
        ref_on_new_axis = reference.to_interpolated_wavelength(calibrated.wavelength)
        A = ref_on_new_axis.spectrum
        B = calibrated.spectrum
        ax[1, 0].plot(calibrated.wavelength_nm(), A - B, label="Residual")
        ax[1, 0].set_ylim(-0.1, 0.1)
        ax[1, 0].set_ylabel("Intensity (Arb. unit)")
        ax[1, 0].set_xlabel("Wavelength (nm)")
        ax[1, 0].legend()
        direct_shifted = direct.to_corrected_wavelength(calibrated.wavelength)
        residual = B - A
        new_weights = (
            direct_shifted.spectrum
            * (calibrated.spectrum - residual)
            / (direct_shifted.spectrum**2 + wiener_noise_level.value)
        )
        new_weights = (
            parametrized_cal.intensity_factors
            - direct_shifted.spectrum
            * residual
            / (direct_shifted.spectrum**2 + wiener_noise_level.value)
        )
        adjusted_cal = aw.data.SpectrometerCalibration(
            intensity_factors=new_weights,
            corrected_frequencies=parametrized_cal.corrected_frequencies,
            corrected_wavelengths=parametrized_cal.corrected_wavelengths,
        )
        second_calibration = adjusted_cal.apply_to_spectrum(direct)
        ax[1, 1].plot(
            second_calibration.wavelength_nm(),
            second_calibration.spectrum,
            label="Calibrated",
        )
        ax[1, 1].plot(reference.wavelength_nm(), reference.spectrum, label="Reference")
        ax[1, 1].set_ylabel("Intensity (Arb. unit)")
        ax[1, 1].set_xlabel("Wavelength (nm)")
        ax[1, 1].legend()
        ax[1, 2].plot(
            adjusted_cal.corrected_wavelengths * 1e9,
            adjusted_cal.intensity_factors,
            label="Final weights",
        )
        ax[1, 2].set_ylabel("Final weights")
        ax[1, 2].set_xlabel("Wavelength (nm)")
        ax[1, 2].legend()
    aw.plot.showmo()
    return (adjusted_cal,)


@app.cell
def _(mo):
    mo.md(r"""If the result is good, save it as an .npz file, and contribute it to the database :)""")
    return


@app.cell
def _(mo):
    save_button = mo.ui.run_button(label="Save calibration")
    save_button
    return (save_button,)


@app.cell
def _(adjusted_cal, filedialog, mo, save_button):
    mo.stop(not save_button.value)

    _file_path = filedialog.asksaveasfilename(
        title="Save File", filetypes=[("npz files", "*.npz")]
    )

    if _file_path is not None:
        try:
            adjusted_cal.save_npz(_file_path)
        except NameError:
            print("Can't save without data")

    return


@app.cell
def _():
    import marimo as mo
    import attoworld as aw
    aw.plot.set_style('nick_dark')
    import matplotlib.pyplot as plt
    import numpy as np
    import scipy
    import tkinter as tk
    from tkinter import filedialog

    root = tk.Tk()
    root.withdraw()
    return aw, filedialog, mo, np, plt, scipy


@app.cell
def _():
    return


if __name__ == "__main__":
    app.run()



import scipy.io as sio
from scipy import constants
import numpy as np
import pandas as pd
from ..numeric import interpolate
from ..spectrum import wavelength_to_frequency
from pathlib import Path
from typing import Optional
from .data_structures import (
    Waveform,
    IntensitySpectrum,
    Spectrogram,
    FrogData,
    ComplexSpectrum,
)


def read_dwc(file_path):
    """
    Reads files in the .dwc format produced by many FROG scanners

    Args:
        file_path: path to the .dwc file

    Returns:
        Spectrogram: the loaded data
    """
    with open(file_path, "r") as file:
        lines = file.readlines()
        delay_increment = float(lines[2].strip().split("=")[1])

    wavelength_vector = pd.read_csv(
        file_path, skiprows=5, nrows=1, delimiter="\t", header=None
    ).values[0]

    data_array = pd.read_csv(
        file_path, skiprows=8, delimiter="\t", header=None, dtype=float
    ).values

    freqs, first_spec = wavelength_to_frequency(wavelength_vector, data_array[:, 0])
    if freqs is not None:
        data_array_freq = np.zeros((len(freqs), data_array.shape[1]))
        data_array_freq[:, 0] = first_spec
        for _i in range(data_array.shape[1]):
            _f, _dat = wavelength_to_frequency(
                wavelength_vector, data_array[:, _i], freqs
            )
            data_array_freq[:, _i] = _dat
        dt = 1e-15 * delay_increment
        delays = dt * np.array(range(data_array.shape[1]))
        delays -= np.mean(delays)
        return Spectrogram(data=data_array_freq, time=delays, freq=freqs)
    else:
        raise Exception("Interpolation failure reading dwc file")


def load_mean_spectrum_from_scarab(filename: str):
    data = np.loadtxt(filename)
    return IntensitySpectrum(
        spectrum=np.mean(data[:, 1::], axis=1),
        wavelength=1e-9 * data[:, 0],
        freq=1e9 * constants.speed_of_light / data[:, 0],
        is_frequency_scaled=False,
        phase=None,
    )


def load_spectrum_from_text(
    filename: str,
    wavelength_multiplier: float = 1e-9,
    wavelength_field: str = "wavelength (nm)",
    spectrum_field: str = "intensity (a.u.)",
    sep: str = "\t",
):
    """
    Load a spectrum contained in a text file.

    Args:
        filename (str): path to the file
        wavelength_multiplier (float): multiplier to convert wavelength to m
        wavelength_field (str): name of the field in the data corresponding to wavelength
        spectrum_field (str): name of the field in the data corresponding to spectral intensity
        sep (str): column separator
    Returns:
        IntensitySpectrum: the intensity spectrum"""
    data = pd.read_csv(filename, sep=sep)
    wavelength = wavelength_multiplier * np.array(data[wavelength_field])
    freq = constants.speed_of_light / wavelength
    spectrum = np.array(data[spectrum_field])
    return IntensitySpectrum(
        spectrum=spectrum,
        wavelength=wavelength,
        freq=freq,
        phase=np.zeros(spectrum.shape, dtype=float),
    )


def load_waveform_from_text(
    filename: str,
    time_multiplier: float = 1e-15,
    time_field: str = "delay (fs)",
    wave_field: str = "field (a.u.)",
    sep="\t",
) -> Waveform:
    """Loads a waveform from a text file

    Args:
        filename (str): path to the file
        time_multiplier (float): multiplier needed to convert the time unit in the file to seconds
        time_field (str): name (header) of the column containing the times
        wave_field (str): name (header) of the column containing the waveform
        sep (str): separator used in the file format (tab is default)

    Returns:
        Waveform: the waveform
    """

    data = pd.read_csv(filename, sep=sep)
    time = time_multiplier * data[time_field].to_numpy()
    wave = data[wave_field].to_numpy()
    dt = time[1] - time[0]
    diff_time = np.diff(time)
    uniform = bool(np.all(np.isclose(diff_time, diff_time[0])))
    return Waveform(wave=wave, time=time, dt=dt, is_uniformly_spaced=uniform)


def load_waves_from_matfile(filename: str, phase: Optional[float] = None):
    """Load the contents of an attolab scanner file in .mat format

    Args:
        phase (float): phase to use when interpreting the lock-in data
        filename (str): path to the mat file
    Returns:
        time_delay: array of time delay values
        signal: signals corresponding to the time delays
    """

    datablob = sio.loadmat(filename)
    stage_position = datablob["xdata"][0, :]
    time_delay = -2e-3 * stage_position / 2.9979e8
    lia_x = datablob["x0"]
    lia_y = datablob["y0"]
    if phase is None:
        optimized_phase = np.atan2(np.sum(lia_y[:] ** 2), np.sum(lia_x[:] ** 2))
        signal = np.fliplr(
            lia_x * np.cos(optimized_phase) + lia_y * np.sin(optimized_phase)
        )
        return time_delay, signal
    else:
        signal = np.fliplr(lia_x * np.cos(phase) - lia_y * np.sin(phase))
        return time_delay, signal


def read_Trebino_FROG_matrix(filename: Path | str) -> Spectrogram:
    """
    Read a spectrogram file made by the Trebino FROG code

    Args:
        filename (Path | str): the name (path) of the file
    """
    with open(filename, "r") as f:
        line = str(f.readline())
        line = line.split()
        n1 = int(line[0])
        n2 = int(line[1])
        line = str(f.readline())
        line = line.split()
    measured_data = pd.read_csv(filename, sep="\t", header=None, skiprows=2)
    measure = []
    raw_freq = (
        1e9 * constants.speed_of_light / np.array(measured_data[0][0:n2]).squeeze()
    )
    df = np.mean(np.diff(raw_freq))
    freq = raw_freq[0] + df * np.array(range(raw_freq.shape[0]))
    time = 1e-15 * np.array(measured_data[0][n2 : (n2 + n1)]).squeeze()
    for i in range(n1):
        measure.append(measured_data[0][(i + 2) * n2 : (i + 3) * n2])
    data = np.array(measure)
    return Spectrogram(data=data, time=time, freq=freq)


def read_Trebino_FROG_speck(filename: Path | str) -> ComplexSpectrum:
    """
    Read a .Speck file made by the Trebino FROG code

    Args:
        filename (Path | str): the name (path) of the file
    """
    data = np.array(pd.read_csv(filename, sep="\t", header=None), dtype=float)
    raw_freq = 1e9 * constants.speed_of_light / data[:, 0]
    df = np.mean(np.diff(raw_freq))
    freq = np.linspace(0.0, raw_freq[-1], int(np.ceil(raw_freq[-1] / df)))
    spectrum = interpolate(freq, raw_freq, data[:, 3]) + 1j * interpolate(
        freq, raw_freq, data[:, 4]
    )
    return ComplexSpectrum(spectrum=spectrum, freq=freq)


def read_Trebino_FROG_data(filename: str) -> FrogData:
    """
    Read a set of data produced by the Trebino FROG reconstruction code.

    Args:
        filename: Base filename of the .bin file; e.g. if the data are mydata.bin.Speck.dat etc., this will be "mydata.bin" """
    spectrum = read_Trebino_FROG_speck(filename + ".Speck.dat")
    pulse = spectrum.to_centered_waveform()
    measured_spectrogram = read_Trebino_FROG_matrix(filename + ".A.dat")
    reconstructed_spectrogram = read_Trebino_FROG_matrix(filename + ".Arecon.dat")
    raw_speck = np.array(
        pd.read_csv(filename + ".Speck.dat", sep="\t", header=None), dtype=float
    )
    raw_ek = np.array(
        pd.read_csv(filename + ".Ek.dat", sep="\t", header=None), dtype=float
    )
    f0 = 1e9 * np.mean(constants.speed_of_light / raw_speck[:, 0])
    dt = 1e-15 * (raw_ek[1, 0] - raw_ek[0, 0])
    raw_reconstruction = raw_ek[:, 3] + 1.0j * raw_ek[:, 4]
    return FrogData(
        spectrum=spectrum,
        pulse=pulse,
        measured_spectrogram=measured_spectrogram,
        reconstructed_spectrogram=reconstructed_spectrogram,
        raw_reconstruction=raw_reconstruction,
        dt=dt,
        f0=float(f0),
    )



"""
This module will contain functions for plotting, and for formatting plots, with the goal of
more easily obtaining a consistent style across plots made by different co-authors.
"""

from .plot import showmo, set_style, label_letter, Char

__all__ = ["showmo", "set_style", "label_letter", "Char"]



from enum import Enum
import importlib.resources


def get_calibration_path():
    with importlib.resources.path(__name__) as data_path:
        return data_path


class CalibrationData(Enum):
    MPQ_ATTO_RESO_MARCO = "MPQ_Atto_Reso_Spectrometer_Marco.npz"


class CalibrationLampReferences(Enum):
    MPQ_ATTO_DEUTERIUM_HALOGEN = "7315273LS-Deuterium-Halogen_CC-VISNIR.lmp"
    MPQ_ATTO_HALOGEN = "7315273LS-Halogen_CC-VISNIR.lmp"
